@@ -14,10 +14,28 @@ import ssl
 
 from pyroute2.netlink.generic import ctrlmsg
 from pyroute2.netlink.generic import nlmsg
+from pyroute2.netlink.generic import NetlinkDecodeError
+from pyroute2.netlink.generic import NetlinkHeaderDecodeError
 from pyroute2.netlink.generic import NETLINK_GENERIC
 from pyroute2.netlink.generic import NETLINK_UNUSED
 
 
+class NetlinkError(Exception):
+    '''
+    Base netlink error
+    '''
+    # FIXME move to a dedicated module
+    pass
+
+
+class NetlinkQueueEmpty(NetlinkError):
+    '''
+    Timeout reached on a message queue polling
+    '''
+    def __init__(self, key):
+        self.key = key
+
+
 def _monkey_handshake(self):
     ##
     #
@@ -80,6 +98,39 @@ IPRCMD_RELOAD = 5
 IPRCMD_ROUTE = 6
 
 
+def _repr_sockets(sockets, mode):
+    '''
+    Represent socket as a text string
+    '''
+    ret = []
+    for i in sockets:
+        url = ''
+        if isinstance(i, netlink_socket):
+            ret.append('netlink://%s' % (i.family))
+            continue
+
+        if i.family == socket.AF_UNIX:
+            url = 'unix'
+        if type(i) == ssl.SSLSocket:
+            if url:
+                url += '+'
+            if i.ssl_version == ssl.PROTOCOL_SSLv3:
+                url += 'ssl'
+            elif i.ssl_version == ssl.PROTOCOL_TLSv1:
+                url += 'tls'
+        if not url:
+            url = 'tcp'
+        if i.family == socket.AF_UNIX:
+            url += '://%s' % (i.getsockname())
+        else:
+            if mode == 'local':
+                url += '://%s:%i' % (i.getsockname())
+            elif mode == 'remote':
+                url += '://%s:%i' % (i.getpeername())
+        ret.append(url)
+    return ret
+
+
 class netlink_error(socket.error):
     def __init__(self, code, msg=None):
         msg = msg or os.strerror(code)
@@ -132,8 +183,16 @@ class marshal(object):
                 self.buf.seek(offset)
                 msg_class = self.msg_map.get(msg_type, nlmsg)
                 msg = msg_class(self.buf)
-                msg.decode()
-                msg['header']['error'] = error
+                try:
+                    msg.decode()
+                    msg['header']['error'] = error
+                except NetlinkHeaderDecodeError as e:
+                    # in the case of header decoding error,
+                    # create an empty message
+                    msg = nlmsg()
+                    msg['header']['error'] = e
+                except NetlinkDecodeError as e:
+                    msg['header']['error'] = e
                 self.fix_message(msg)
                 offset += msg.length
                 result.append(msg)
@@ -167,21 +226,36 @@ class ssl_credentials(object):
         self.ca = ca
 
 
+class masq_record(object):
+    def __init__(self, seq, pid, socket):
+        self.seq = seq
+        self.pid = pid
+        self.socket = socket
+        self.ctime = time.time()
+
+    def __repr__(self):
+        return "%s, %s, %s, %s" % (_repr_sockets([self.socket], 'remote'),
+                                   self.pid, self.seq, time.ctime(self.ctime))
+
+
 class iothread(threading.Thread):
     def __init__(self):
         threading.Thread.__init__(self)
         self.setDaemon(True)
+        self.pid = os.getpid()
+        self._nonce = 0
+        self._stop = False
+        self._reload_event = threading.Event()
         # fd lists for select()
         self._rlist = set()
         self._wlist = set()
         self._xlist = set()
-        self._stop = False
         # routing
         self.rtable = {}          # {client_socket: send_method, ...}
         self.families = {}        # {family_id: send_method, ...}
         self.marshals = {}        # {netlink_socket: marshal, ...}
         self.listeners = {}       # {int: Queue(), int: Queue()...}
-        self.netlinks = set()     # set(socket, socket...)
+        self.masquerade = {}      # {int: masq_record()...}
         self.clients = set()      # set(socket, socket...)
         self.uplinks = set()      # set(socket, socket...)
         self.servers = set()      # set(socket, socket...)
@@ -192,10 +266,65 @@ class iothread(threading.Thread):
         self.control = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
         self.control.bind(b'\0%s' % (self.uuid))
         self._rlist.add(self.control)
+        # masquerade cache expiration
+        self._expire_thread = threading.Thread(target=self._expire_masq)
+        self._expire_thread.setDaemon(True)
+        self._expire_thread.start()
+        # buffers reassembling
+        self.buffers = Queue.Queue()
+        self._feed_thread = threading.Thread(target=self._feed_buffers)
+        self._feed_thread.setDaemon(True)
+        self._feed_thread.start()
         # debug
         self.record = False
         self.backlog = []
 
+    def _feed_buffers(self):
+        '''
+        Beckground thread to feed reassembled buffers to the parser
+        '''
+        save = None
+        while not self._stop:
+            (buf, marshal, sock) = self.buffers.get()
+            if save is not None:
+                # concatenate buffers
+                buf.seek(0)
+                save.write(buf.read())
+                save.length += buf.length
+                # discard save
+                buf = save
+                save = None
+            offset = 0
+            while offset < buf.length:
+                buf.seek(offset)
+                length = struct.unpack('I', buf.read(4))[0]
+                if offset + length > buf.length:
+                    # create save buffer
+                    buf.seek(offset)
+                    save = io.BytesIO()
+                    save.length = save.write(buf.read())
+                    # truncate the buffer
+                    buf.truncate(offset)
+                    break
+                offset += length
+            # feed buffer to the parser
+            try:
+                self.parse(buf.getvalue(), marshal, sock)
+            except:
+                traceback.print_exc()
+
+    def _expire_masq(self):
+        '''
+        Background thread that expires masquerade cache entries
+        '''
+        while not self._stop:
+            # expire masquerade records
+            ts = time.time()
+            for i in tuple(self.masquerade.keys()):
+                if (ts - self.masquerade[i].ctime) > 60:
+                    del self.masquerade[i]
+            time.sleep(60)
+
     def _get_socket(self, url, server_side):
         assert type(url) is str
         target = urlparse.urlparse(url)
@@ -234,20 +363,65 @@ class iothread(threading.Thread):
                                    ssl_version=ssl_version)
         return (sock, address)
 
+    def nonce(self):
+        if self._nonce == 0xffffffff:
+            self._nonce = 1
+        else:
+            self._nonce += 1
+        return self._nonce
+
+    def distribute(self, data):
+        """
+        Send message to all clients. Called from self.route()
+        """
+        data.seek(8)
+        # read sequence number and pid
+        seq, pid = struct.unpack('II', data.read(8))
+        # default target -- broadcast
+        target = None
+        # if it is a unicast response
+        if pid == self.pid:
+            # lookup masquerade table
+            target = self.masquerade.get(seq, None)
+        # there is valid masquerade record
+        if target is not None:
+            # fill up client's pid and seq
+            offset = 0
+            # ... but -- for each message in the packet :)
+            while offset < data.length:
+                # write the data
+                data.seek(offset + 8)
+                data.write(struct.pack('II', target.seq, target.pid))
+                # skip to the next message
+                data.seek(offset)
+                length = struct.unpack('I', data.read(4))[0]
+                offset += length
+
+            target.socket.send(data.getvalue())
+        else:
+            # otherwise, broadcast packet
+            for sock in self.clients:
+                sock.send(data.getvalue())
+            # return True -- this packet should be parsed
+            return True
+
     def route(self, sock, data):
+        """
+        Route message
+        """
+        data.seek(4)
         # message type, offset 4 bytes, length 2 bytes
-        mtype = struct.unpack('H', data[4:6])[0]
-        # use NETLINK_UNUSED as inter-pyroute2
+        mtype = struct.unpack('H', data.read(2))[0]
         # FIXME log routing failures
-        if mtype == NETLINK_UNUSED:
-            #
-            # Control messages
-            #
+
+        ##
+        #
+        # NETLINK_UNUSED as inter-pyroute2
+        #
+        if (mtype == NETLINK_UNUSED) and (sock in self.clients):
             cmd = self.parse_control(data)
             if cmd['cmd'] == IPRCMD_ROUTE:
-                #
-                # Route request
-                #
+                # routing request
                 family = cmd.get_attr('CTRL_ATTR_FAMILY_ID')[0]
                 if family in self.families:
                     send = self.families[family]
@@ -256,23 +430,47 @@ class iothread(threading.Thread):
                 # * subscribe requests
                 # * ...
 
+        ##
+        #
+        # NETLINK_UNUSED as intra-pyroute2
+        #
+        elif (mtype == NETLINK_UNUSED) and (sock == self.control):
+            cmd = self.parse_control(data)
+            if cmd['cmd'] == IPRCMD_STOP:
+                # Stop iothread
+                self._stop = True
+            elif cmd['cmd'] == IPRCMD_RELOAD:
+                # Reload io cycle
+                self._reload_event.set()
+
+        ##
+        #
+        # Data messages
+        #
         else:
-            #
-            # Data messages
-            #
+            if sock in self.clients:
+                # create masquerade record for client's messages
+                # 1. generate nonce
+                nonce = self.nonce()
+                # 2. save masquerade record, invalidating old one
+                data.seek(8)
+                seq, pid = struct.unpack('II', data.read(8))
+                self.masquerade[nonce] = masq_record(seq, pid, sock)
+                # 3. overwrite seq and pid
+                data.seek(8)
+                data.write(struct.pack('II', nonce, self.pid))
+
             if sock in self.rtable:
-                send = self.rtable[sock]
-                send(data)
+                if self.rtable[sock](data):
+                    self.buffers.put((data, self.marshals[sock], sock))
 
     def parse_control(self, data):
-        buf = io.BytesIO()
-        buf.write(data)
-        buf.seek(0)
-        cmd = ctrlmsg(buf)
+        data.seek(0)
+        cmd = ctrlmsg(data)
         cmd.decode()
         return cmd
 
-    def parse(self, data, marshal):
+    def parse(self, data, marshal, sock):
         '''
         Parse and enqueue messages. A message can be
         retrieved from netlink socket as well as from a
@@ -293,6 +491,7 @@ class iothread(threading.Thread):
             if self.record:
                 self.backlog.append((time.asctime(), msg))
             key = msg['header']['sequence_number']
+            msg['header']['host'] = _repr_sockets([sock], 'remote')[0]
             if key not in self.listeners:
                 key = 0
             if self.mirror and key != 0:
@@ -312,29 +511,14 @@ class iothread(threading.Thread):
         return self.command(IPRCMD_STOP)
 
     def reload(self):
-        return self.command(IPRCMD_RELOAD)
-
-    def add_netlink(self, family, groups):
         '''
-        Add netlink connection and reload I/O cycle.
+        Reload I/O cycle. Warning: this method should be never
+        called from iothread, as it will not return.
         '''
-        sock = netlink_socket(family)
-        sock.bind(groups)
-        self._rlist.add(sock)
-        self.netlinks.add(sock)
-        self.reload()
-        return sock
-
-    def remove_netlink(self, family, groups):
-        sock = None
-        for sock in self.netlinks:
-            if (sock.family, sock.groups) == (family, groups):
-                break
-        assert sock
-        self.netlinks.remove(sock)
-        self._rlist.remove(sock)
-        sock.close()
-        return sock
+        self._reload_event.clear()
+        ret = self.command(IPRCMD_RELOAD)
+        self._reload_event.wait()
+        return ret
 
     def add_server(self, url):
         '''
@@ -360,24 +544,30 @@ class iothread(threading.Thread):
         self.reload()
         return sock
 
-    def add_uplink(self, url):
+    def add_uplink(self, url=None, family=None, groups=None):
         '''
-        Add an uplink server to get information from
+        Add an uplink to get information from:
+
+        * url -- remote serve
+        * family and groups -- netlink
         '''
-        (sock, address) = self._get_socket(url, server_side=False)
-        sock.connect(address)
+        if url is not None:
+            (sock, address) = self._get_socket(url, server_side=False)
+            sock.connect(address)
+        elif (family is not None) and \
+             (groups is not None):
+            sock = netlink_socket(family)
+            sock.bind(groups)
+        else:
+            raise Exception("uplink type not supported")
         self._rlist.add(sock)
         self.uplinks.add(sock)
+        self.rtable[sock] = self.distribute
         self.reload()
         return sock
 
-    def remove_uplink(self, sock=None, address=None):
-        assert sock or address
-        if address:
-            for sock in self.uplinks:
-                if sock.getsockname() == address:
-                    break
-        assert sock
+    def remove_uplink(self, sock):
+        assert sock in self.uplinks
         self._rlist.remove(sock)
         self.uplinks.remove(sock)
         self.reload()
@@ -391,7 +581,6 @@ class iothread(threading.Thread):
         self._rlist.add(sock)
         self._wlist.add(sock)
         self.clients.add(sock)
-        self.reload()
         return sock
 
     def remove_client(self, sock):
@@ -425,64 +614,29 @@ class iothread(threading.Thread):
                 # Receive data from already open connection
                 #
                 # FIXME max length
-                data = fd.recv(16384)
-                if self.record:
-                    self.backlog.append((time.asctime(), data))
+                data = io.BytesIO()
+                data.length = data.write(fd.recv(16384))
 
                 ##
                 #
-                # Control socket, only type == NETLINK_UNUSED
+                # Close socket
                 #
-                if fd == self.control:
-                    # FIXME move it to specific marshal
-                    cmd = self.parse_control(data)
-                    if cmd['cmd'] == IPRCMD_STOP:
-                        self._stop = True
-                        break
-                    elif cmd['cmd'] == IPRCMD_RELOAD:
-                        pass
-
-                ##
-                #
-                # Incoming packet from local netlink
-                #
-                elif fd in self.netlinks:
-                    # retranslate to all clients
-                    # FIXME what about subscriptions?
-                    for sock in self.clients:
-                        sock.send(data)
-                    # parse message locally
-                    self.parse(data, self.marshals[fd])
-
-                ##
-                #
-                # Incoming packet from remote client
-                #
-                elif fd in self.clients:
-                    if data == '':
-                        # client closed connection
+                if data.length == 0:
+                    if fd in self.clients:
                         self.remove_client(fd)
-                    else:
-                        try:
-                            self.route(fd, data)
-                        except:
-                            # drop packet on any error
-                            traceback.print_exc()
+                    elif fd in self.uplinks:
+                        self.remove_uplink(fd)
+                    continue
 
                 ##
                 #
-                # Incoming packet from remote uplink
+                # Route the data
                 #
-                elif fd in self.uplinks:
-                    if data == '':
-                        # uplink closed connection
-                        self.remove_uplink(fd)
-                    else:
-                        try:
-                            self.parse(data, self.marshals[fd])
-                        except:
-                            # drop packet on any error
-                            pass
+                if self.record:
+                    self.backlog.append((time.asctime(),
+                                         fd.getsockname(),
+                                         data))
+                self.route(fd, data)
 
 
 class netlink(object):
@@ -491,61 +645,58 @@ class netlink(object):
     to monitor network and netlink I/O, creates and destroys message
     queues.
 
-    It can operate in three modes:
-     * local system only
-     * server mode
-     * client mode
+    By default, netlink class connects to the local netlink socket
+    on startup. If you prefer to connect to another host, use::
 
-    By default, it starts in local system only mode. To start a
-    server, you should call netlink.serve(url). The method
-    can be called several times to listen on specific interfaces
-    and/or ports.
+        nl = netlink(host='tcp://remote.01host:7000')
 
-    Alternatively, you can start the object in the client mode.
-    In that case you should provide server url in the host
-    parameter. You can not mix server and client modes, so
-    message proxy/relay not possible yet. This will be fixed
-    in the future.
+    It is possible to connect to uplinks after the startup::
 
-    Urls should be specified in the form:
-        (host, port)
+        nl = netlink(do_connect=False)
+        nl.connect('tcp://remote.01host:7000')
 
-    E.g.:
-        nl = netlink(host=('127.0.0.1', 7000))
+    To act as a server, call serve()::
+
+        nl = netlink(do_connect=False)
+        nl.connect('localsystem')
+        nl.serve('unix:///tmp/pyroute')
     '''
 
     family = NETLINK_GENERIC
     groups = 0
     marshal = marshal
 
-    def __init__(self, debug=False, host='localsystem', interruptible=False,
-                 key=None, cert=None, ca=None):
-        self.server = host
+    def __init__(self, debug=False, timeout=3, do_connect=True,
+                 host='localsystem', key=None, cert=None, ca=None):
+        self._timeout = timeout
         self.iothread = iothread()
         self.listeners = self.iothread.listeners
         self.ssl_keys = self.iothread.ssl_keys
-        self.interruptible = interruptible
-        if key:
-            self.ssl_keys[host] = ssl_credentials(key, cert, ca)
-        if host == 'localsystem':
-            self.socket = self.iothread.add_netlink(self.family, self.groups)
-        else:
-            self.socket = self.iothread.add_uplink(self.server)
-            self.request_route()
-        self.iothread.marshals[self.socket] = self.marshal()
+        self._sockets = set()
+        self.servers = {}
         self.iothread.families[self.family] = self.send
         self.iothread.start()
         self.debug = debug
-        self._nonce = 1
-        self.servers = {}
+        if do_connect:
+            self.connect(host, key, cert, ca)
 
-    def request_route(self):
-        rmsg = ctrlmsg()
-        rmsg['header']['type'] = NETLINK_UNUSED
-        rmsg['cmd'] = IPRCMD_ROUTE
-        rmsg['attrs'] = (('CTRL_ATTR_FAMILY_ID', self.family), )
-        rmsg.encode()
-        self.socket.send(rmsg.buf.getvalue())
+    def connect(self, host='localsystem', key=None, cert=None, ca=None):
+        if key:
+            self.ssl_keys[host] = ssl_credentials(key, cert, ca)
+        if host == 'localsystem':
+            sock = self.iothread.add_uplink(family=self.family,
+                                            groups=self.groups)
+        else:
+            sock = self.iothread.add_uplink(url=host)
+            smsg = ctrlmsg()
+            smsg['header']['type'] = NETLINK_UNUSED
+            smsg['header']['pid'] = os.getpid()
+            smsg['cmd'] = IPRCMD_ROUTE
+            smsg['attrs'] = (('CTRL_ATTR_FAMILY_ID', self.family), )
+            smsg.encode()
+            sock.send(smsg.buf.getvalue())
+        self.iothread.marshals[sock] = self.marshal()
+        self._sockets.add(sock)
 
     def shutdown(self, url=None):
         url = url or self.servers.keys()[0]
@@ -554,22 +705,20 @@ class netlink(object):
             del self.ssl_keys[url]
         del self.servers[url]
 
+    def get_servers(self):
+        return _repr_sockets(self.iothread.servers, 'local')
+
+    def get_clients(self):
+        return _repr_sockets(self.iothread.clients, 'remote')
+
+    def get_sockets(self):
+        return _repr_sockets(self._sockets, 'remote')
+
     def serve(self, url, key=None, cert=None, ca=None):
         if key:
             self.ssl_keys[url] = ssl_credentials(key, cert, ca)
         self.servers[url] = self.iothread.add_server(url)
 
-    def nonce(self):
-        '''
-        Increment netlink protocol nonce (there is no need to
-        call it directly)
-        '''
-        if self._nonce == 0xffffffff:
-            self._nonce = 1
-        else:
-            self._nonce += 1
-        return self._nonce
-
     def mirror(self, operate=True):
         '''
         Turn message mirroring on/off. When it is 'on', all
@@ -594,45 +743,50 @@ class netlink(object):
         else:
             del self.listeners[0]
 
-    def get(self, key=0, interruptible=False):
+    def _remove_queue(self, key):
         '''
-        Get a message from a queue
-
-        * key -- message queue number
-        * interruptible -- catch ctrl-c
-
-        Please note, that setting interruptible=True will cause
-        polling overhead. Python starts implied poll cycle, if
-        timeout is set.
+        Flush the queue to the default one and remove it
         '''
         queue = self.listeners[key]
-        interruptible = interruptible or self.interruptible
-        if interruptible:
-            tot = 31536000
-        else:
-            tot = None
-        result = []
-        while True:
-            # timeout is set to catch ctrl-c
-            # Bug-Url: http://bugs.python.org/issue1360
-            msg = queue.get(block=True, timeout=tot)
-            if msg['header']['error'] is not None:
-                raise msg['header']['error']
-            if msg['header']['type'] != NLMSG_DONE:
-                result.append(msg)
-            if (msg['header']['type'] == NLMSG_DONE) or \
-               (not msg['header']['flags'] & NLM_F_MULTI):
-                break
-        # not default queue
+        # only not the default queue
         if key != 0:
             # delete the queue
             del self.listeners[key]
             # get remaining messages from the queue and
             # re-route them to queue 0 or drop
             while not queue.empty():
-                msg = queue.get(bloc=True, timeout=tot)
+                msg = queue.get()
                 if 0 in self.listeners:
                     self.listeners[0].put(msg)
+
+    def get(self, key=0):
+        '''
+        Get a message from a queue
+
+        * key -- message queue number
+        '''
+        queue = self.listeners[key]
+        result = []
+        hosts = len(self._sockets)
+        while True:
+            # timeout should alse be set to catch ctrl-c
+            # Bug-Url: http://bugs.python.org/issue1360
+            try:
+                msg = queue.get(block=True, timeout=self._timeout)
+            except Queue.Empty:
+                self._remove_queue(key)
+                raise NetlinkQueueEmpty(key)
+            if msg['header']['error'] is not None:
+                self._remove_queue(key)
+                raise msg['header']['error']
+            if msg['header']['type'] != NLMSG_DONE:
+                result.append(msg)
+            if (msg['header']['type'] == NLMSG_DONE) or \
+               (not msg['header']['flags'] & NLM_F_MULTI):
+                hosts -= 1
+            if hosts == 0:
+                break
+        self._remove_queue(key)
         return result
 
     def send(self, buf):
@@ -640,10 +794,12 @@ class netlink(object):
         Send a buffer or to the local kernel, or to
         the server, depending on the setup.
         '''
-        if self.server == 'localsystem':
-            self.socket.sendto(buf, (0, 0))
-        else:
-            self.socket.send(buf)
+        data = buf.getvalue()
+        for sock in self._sockets:
+            if isinstance(sock, netlink_socket):
+                sock.sendto(data, (0, 0))
+            else:
+                sock.send(data)
 
     def nlm_request(self, msg, msg_type,
                     msg_flags=NLM_F_DUMP | NLM_F_REQUEST):
@@ -652,14 +808,14 @@ class netlink(object):
         fields, and wait for response.
         '''
         # FIXME make it thread safe, yeah
-        nonce = self.nonce()
+        nonce = self.iothread.nonce()
         self.listeners[nonce] = Queue.Queue()
         msg['header']['sequence_number'] = nonce
         msg['header']['pid'] = os.getpid()
         msg['header']['type'] = msg_type
         msg['header']['flags'] = msg_flags
         msg.encode()
-        self.send(msg.buf.getvalue())
+        self.send(msg.buf)
         result = self.get(nonce)
         if not self.debug:
             for i in result:
