@@ -1,6 +1,7 @@
 
 import socket
 import struct
+import types
 import io
 
 from pyroute2.common import hexdump
@@ -31,6 +32,37 @@ NETLINK_SCSITRANSPORT = 18   # SCSI Transports
 NLMSG_ALIGNTO = 4
 
 
+class NetlinkDecodeError(Exception):
+    '''
+    Base decode error class.
+
+    Incapsulates underlying error for the following analysis
+    '''
+    def __init__(self, exception):
+        self.exception = exception
+
+
+class NetlinkHeaderDecodeError(NetlinkDecodeError):
+    '''
+    The error occured while decoding a header
+    '''
+    pass
+
+
+class NetlinkDataDecodeError(NetlinkDecodeError):
+    '''
+    The error occured while decoding the message fields
+    '''
+    pass
+
+
+class NetlinkNLADecodeError(NetlinkDecodeError):
+    '''
+    The error occured while decoding NLA chain
+    '''
+    pass
+
+
 def NLMSG_ALIGN(l):
     return (l + NLMSG_ALIGNTO - 1) & ~ (NLMSG_ALIGNTO - 1)
 
@@ -160,24 +192,34 @@ class nlmsg_base(dict):
 
     def decode(self):
         self.offset = self.buf.tell()
-        # read the header
+        # decode the header
         if self.header is not None:
-            self['header'].decode()
-            # update length from header
-            # it can not be less than 4
-            self.length = max(self['header']['length'], 4)
-            if self.fmt == 's':
-                self.fmt = '%is' % (self.length - 4)
-            elif self.fmt == 'z':
-                self.fmt = '%is' % (self.length - 5)
-        # read the data
-        size = struct.calcsize(self.fmt)
-        self.update(dict(zip(self.fields,
-                         struct.unpack(self.fmt, self.buf.read(size)))))
-        # align NLA chain start
-        self.buf.seek(NLMSG_ALIGN(self.buf.tell()))
-        # read NLA chain
-        self.decode_nlas()
+            try:
+                self['header'].decode()
+                # update length from header
+                # it can not be less than 4
+                self.length = max(self['header']['length'], 4)
+                if self.fmt == 's':
+                    self.fmt = '%is' % (self.length - 4)
+                elif self.fmt == 'z':
+                    self.fmt = '%is' % (self.length - 5)
+            except Exception as e:
+                raise NetlinkHeaderDecodeError(e)
+        # decode the data
+        try:
+            size = struct.calcsize(self.fmt)
+            self.update(dict(zip(self.fields,
+                             struct.unpack(self.fmt, self.buf.read(size)))))
+        except Exception as e:
+            raise NetlinkDataDecodeError(e)
+        # decode NLA
+        try:
+            # align NLA chain start
+            self.buf.seek(NLMSG_ALIGN(self.buf.tell()))
+            # read NLA chain
+            self.decode_nlas()
+        except Exception as e:
+            raise NetlinkNLADecodeError(e)
         if len(self['attrs']) == 0:
             del self['attrs']
         if self['value'] is NotInitialized:
@@ -263,12 +305,12 @@ class nlmsg_base(dict):
         self.r_nla_map = {}
 
         # create enumeration
-        types = enumerate((i[0] for i in self.nla_map))
+        nla_types = enumerate((i[0] for i in self.nla_map))
         # that's a little bit tricky, but to reduce
         # the required amount of code in modules, we have
         # to jump over the head
         zipped = [(k[1][0], k[0][0], k[0][1]) for k in
-                  zip(self.nla_map, types)]
+                  zip(self.nla_map, nla_types)]
 
         for (key, name, nla_class) in zipped:
             # lookup NLA class
@@ -307,6 +349,10 @@ class nlmsg_base(dict):
             if msg_type in self.t_nla_map:
                 # get the class
                 msg_class = self.t_nla_map[msg_type][0]
+                # is it a class or a function?
+                if isinstance(msg_class, types.MethodType):
+                    # if it is a function -- use it to get the class
+                    msg_class = msg_class()
                 # and the name
                 msg_name = self.t_nla_map[msg_type][1]
                 try:
@@ -450,4 +496,9 @@ class ctrlmsg(genlmsg):
     # FIXME: to be extended
     nla_map = (('CTRL_ATTR_UNSPEC', 'none'),
                ('CTRL_ATTR_FAMILY_ID', 'uint16'),
-               ('CTRL_ATTR_FAMILY_NAME', 'asciiz'))
+               ('CTRL_ATTR_FAMILY_NAME', 'asciiz'),
+               ('CTRL_ATTR_VERSION', 'hex'),
+               ('CTRL_ATTR_HDRSIZE', 'hex'),
+               ('CTRL_ATTR_MAXATTR', 'hex'),
+               ('CTRL_ATTR_OPS', 'hex'),
+               ('CTRL_ATTR_MCAST_GROUPS', 'hex'))
